@@ -1,14 +1,18 @@
 # This relies on each of the submodules having an __all__ variable.
 
+__version__ = '0.7.3'
+
+
 from .protocol import *
+from .connector import *
 from .client import *
 from .errors import *
 from .parsers import *
-from .session import *
 
 
 __all__ = (client.__all__ +
            errors.__all__ +
            parsers.__all__ +
            protocol.__all__ +
-           session.__all__)
+           connector.__all__ +
+           ['__version__'])
