@@ -6,7 +6,6 @@ import asyncio
 import base64
 import collections
 import email.message
-import functools
 import http.client
 import http.cookies
 import json
@@ -17,11 +16,11 @@ import logging
 import mimetypes
 import os
 import random
-import ssl
-import socket
 import time
 import uuid
 import urllib.parse
+import weakref
+import warnings
 
 import aiohttp
 
@@ -39,50 +38,49 @@ def request(method, url, *,
             encoding='utf-8',
             version=(1, 1),
             timeout=None,
-            conn_timeout=None,
             compress=None,
             chunked=None,
             expect100=False,
-            session=None,
-            verify_ssl=True,
-            connection_params=None,
+            connector=None,
             loop=None,
+            read_until_eof=True,
             request_class=None):
     """Constructs and sends a request. Returns response object.
 
-    method: http method
-    url: request url
-    params: (optional) Dictionary or bytes to be sent in the query string
-      of the new request
-    data: (optional) Dictionary, bytes, or file-like object to
+    :param method: http method
+    :param url: request url
+    :param params: (optional) Dictionary or bytes to be sent in the query
+      string of the new request
+    :param data: (optional) Dictionary, bytes, or file-like object to
       send in the body of the request
-    headers: (optional) Dictionary of HTTP Headers to send with the request
-    cookies: (optional) Dict object to send with the request
-    files: (optional) Dictionary of 'name': file-like-objects
+    :param headers: (optional) Dictionary of HTTP Headers to send with
+      the request
+    :param cookies: (optional) Dict object to send with the request
+    :param files: (optional) Dictionary of 'name': file-like-objects
        for multipart encoding upload
-    auth: (optional) Auth tuple to enable Basic HTTP Auth
-    timeout: (optional) Float describing the timeout of the request
-    conn_timeout: (optional) Float describing the timeout
-       of the host connection
-    allow_redirects: (optional) Boolean. Set to True if POST/PUT/DELETE
+    :param auth: (optional) Auth tuple to enable Basic HTTP Auth
+    :param timeout: (optional) Float describing the timeout of the request
+    :param allow_redirects: (optional) Boolean. Set to True if POST/PUT/DELETE
        redirect following is allowed.
-    compress: Boolean. Set to True if request has to be compressed
+    :param compress: Boolean. Set to True if request has to be compressed
        with deflate encoding.
-    chunked: Boolean or Integer. Set to chunk size for chunked
+    :param chunked: Boolean or Integer. Set to chunk size for chunked
        transfer encoding.
-    expect100: Boolean. Expect 100-continue response from server.
-    session: aiohttp.Session instance to support connection pooling and
-       session cookies.
-    loop: Optional event loop.
+    :param expect100: Boolean. Expect 100-continue response from server.
+    :param connector: aiohttp.conntect.BaseConnector instance to support
+       connection pooling and session cookies.
+    :param read_until_eof: Read response until eof if response
+       does not have Content-Length header.
+    :param request_class: Custom Request class implementation.
+    :param loop: Optional event loop.
 
-    Usage:
+    Usage::
 
-      import aiohttp
-      >> resp = yield from aiohttp.request('GET', 'http://python.org/')
-      >> resp
+      >>> import aiohttp
+      >>> resp = yield from aiohttp.request('GET', 'http://python.org/')
+      >>> resp
       <HttpResponse(python.org/) [200]>
-
-      >> data = yield from resp.read()
+      >>> data = yield from resp.read_and_close()
 
     """
     redirects = 0
@@ -90,40 +88,37 @@ def request(method, url, *,
         loop = asyncio.get_event_loop()
     if request_class is None:
         request_class = HttpRequest
+    if connector is None:
+        connector = aiohttp.TCPConnector(loop=loop)
 
     while True:
+        using_proxy = isinstance(connector, aiohttp.ProxyConnector)
         req = request_class(
             method, url, params=params, headers=headers, data=data,
             cookies=cookies, files=files, auth=auth, encoding=encoding,
             version=version, compress=compress, chunked=chunked,
-            verify_ssl=verify_ssl, loop=loop, expect100=expect100)
+            loop=loop, expect100=expect100, using_proxy=using_proxy)
 
-        if session is None:
-            conn = _connect(req, loop, connection_params)
-        else:
-            conn = session.start(req, loop, connection_params)
-
-        if conn_timeout is None and timeout is not None:
-            conn_timeout = timeout
-
-        conn_task = asyncio.async(conn, loop=loop)
         try:
-            if conn_timeout:
-                transport, proto, wrp = yield from asyncio.wait_for(
-                    conn_task, conn_timeout, loop=loop)
-            else:
-                transport, proto, wrp = yield from conn_task
+            conn = yield from connector.connect(req)
 
-            resp = yield from _make_request(
-                transport, proto, req, wrp, timeout, loop)
+            resp = req.send(conn.writer, conn.reader)
+            try:
+                if timeout:
+                    yield from asyncio.wait_for(
+                        resp.start(conn, read_until_eof), timeout, loop=loop)
+                else:
+                    yield from resp.start(conn, read_until_eof)
+            except:
+                resp.close()
+                conn.close()
+                raise
         except asyncio.TimeoutError:
             raise aiohttp.TimeoutError from None
         except aiohttp.BadStatusLine as exc:
             raise aiohttp.ClientConnectionError(exc)
         except OSError as exc:
             raise aiohttp.OsConnectionError(exc)
-        finally:
-            conn_task.cancel()
 
         # redirects
         if resp.status in (301, 302) and allow_redirects:
@@ -150,220 +145,6 @@ def request(method, url, *,
     return resp
 
 
-@asyncio.coroutine
-def _connect(req, loop, params):
-    if params is not None:
-        transport, proto = yield from loop.create_connection(
-            functools.partial(aiohttp.StreamProtocol, loop=loop),
-            params['host'], params['port'],
-            ssl=params['ssl'], family=params['family'],
-            proto=params['proto'], flags=params['flags'])
-    else:
-        transport, proto = yield from loop.create_connection(
-            functools.partial(aiohttp.StreamProtocol, loop=loop),
-            req.host, req.port, ssl=req.ssl)
-    wrp = TransportWrapper(transport)
-    return transport, proto, wrp
-
-
-@asyncio.coroutine
-def _make_request(transport, proto, req, wrapper, timeout, loop):
-    try:
-        resp = req.send(transport)
-        if timeout:
-            yield from asyncio.wait_for(
-                resp.start(proto, wrapper), timeout, loop=loop)
-        else:
-            yield from resp.start(proto, wrapper)
-    except:
-        transport.close()
-        raise
-    else:
-        return resp
-
-
-class TransportWrapper:
-
-    def __init__(self, transport):
-        self.transport = transport
-
-    def close(self, force=False):
-        self.transport.close()
-
-
-class HttpClient:
-    """Allow to use mutiple hosts with same path. And automatically
-    mark failed hosts.
-    """
-
-    _resolve_timeout = 360.0  # update dns info every 5 minutes
-
-    def __init__(self, hosts, *,
-                 method=None, path=None, ssl=False, session=False,
-                 timeout=None, conn_timeout=None, failed_timeout=5.0,
-                 resolve=True, loop=None):
-        super().__init__()
-
-        if isinstance(hosts, str):
-            hosts = (hosts,)
-
-        if not hosts:
-            raise ValueError('Hosts are required')
-
-        self._hosts = []
-        for host in hosts:
-            if isinstance(host, str):
-                if ':' in host:
-                    host, port = host.split(':')
-                    try:
-                        port = int(port)
-                    except:
-                        raise ValueError('Port has to be integer: %s' % host)
-                else:
-                    port = 80
-            else:
-                host, port = host
-
-            self._hosts.append((host, port))
-
-        self._method = method
-        self._path = path
-        self._ssl = ssl
-        self._timeout = timeout
-        self._conn_timeout = conn_timeout
-        self._schema = 'https' if ssl else 'http'
-        self._session = aiohttp.Session() if session else None
-
-        self._failed = collections.deque()
-        self._failed_handle = None
-        self._failed_timeout = failed_timeout
-
-        if loop is None:
-            loop = asyncio.get_event_loop()
-        self._loop = loop
-
-        self._resolve = resolve
-        self._resolved_hosts = {}
-        if resolve:
-            self._resolve_handle = self._loop.call_later(
-                self._resolve_timeout, self._cleanup_resolved_host)
-
-    def _cleanup_resolved_host(self):
-        self._resolved_hosts.clear()
-        self._resolve_handle = self._loop.call_later(
-            self._resolve_timeout, self._cleanup_resolved_host)
-
-    @asyncio.coroutine
-    def _resolve_host(self, host, port):
-        if self._resolve:
-            key = (host, port)
-            if key not in self._resolved_hosts:
-                infos = yield from self._loop.getaddrinfo(
-                    host, port, type=socket.SOCK_STREAM, family=socket.AF_INET)
-
-                hosts = []
-                for family, _, proto, _, address in infos:
-                    hosts.append(
-                        {'host': address[0], 'port': address[1],
-                         'ssl': self._ssl, 'family': family,
-                         'proto': proto, 'flags': socket.AI_NUMERICHOST})
-                self._resolved_hosts[key] = hosts
-
-            return self._resolved_hosts[key]
-        else:
-            return [{'host': host, 'port': port,
-                     'ssl': self._ssl, 'family': 0, 'proto': 0, 'flags': 0}]
-
-    def _resurrect_failed(self):
-        now = int(time.time())
-
-        while self._failed:
-            if (now - self._failed[0][1]) >= self._failed_timeout:
-                self._hosts.append(self._failed.popleft()[0])
-            else:
-                break
-
-        if self._failed:
-            self._failed_handle = self._loop.call_later(
-                self._failed_timeout, self._resurrect_failed)
-        else:
-            self._failed_handle = None
-
-    @asyncio.coroutine
-    def request(self, method=None, path=None, *,
-                params=None,
-                data=None,
-                headers=None,
-                cookies=None,
-                files=None,
-                auth=None,
-                allow_redirects=True,
-                max_redirects=10,
-                encoding='utf-8',
-                version=(1, 1),
-                compress=None,
-                timeout=None,
-                conn_timeout=None,
-                chunked=None,
-                expect100=False,
-                verify_ssl=True):
-
-        if method is None:
-            method = self._method
-        if path is None:
-            path = self._path
-        if timeout is None:
-            timeout = self._timeout
-        if conn_timeout is None:
-            conn_timeout = self._conn_timeout
-
-        # if all hosts marked as failed try first from failed
-        if not self._hosts:
-            self._hosts.append(self._failed.popleft()[0])
-
-        hosts = self._hosts
-
-        while hosts:
-            idx = random.randint(0, len(hosts)-1)
-
-            h_info = hosts[idx]
-            url = urllib.parse.urljoin(
-                '{}://{}:{}'.format(
-                    self._schema, h_info[0], h_info[1]), path)
-            host_params = yield from self._resolve_host(h_info[0], h_info[1])
-
-            for conn_params in host_params:
-                try:
-                    resp = yield from request(
-                        method, url, params=params, data=data, headers=headers,
-                        cookies=cookies, files=files, auth=auth,
-                        encoding=encoding, allow_redirects=allow_redirects,
-                        version=version, max_redirects=max_redirects,
-                        conn_timeout=conn_timeout, timeout=timeout,
-                        compress=compress, chunked=chunked,
-                        verify_ssl=verify_ssl, expect100=expect100,
-                        session=self._session, connection_params=conn_params,
-                        loop=self._loop)
-                except (aiohttp.ConnectionError, aiohttp.TimeoutError):
-                    pass
-                else:
-                    return resp
-
-            if h_info in hosts:
-                # could be removed concurrently
-                hosts.remove(h_info)
-                self._failed.append((h_info, int(time.time())))
-                if not self._failed_handle:
-                    self._failed_handle = self._loop.call_later(
-                        self._failed_timeout, self._resurrect_failed)
-
-                key = (h_info[0], h_info[1])
-                if key in self._resolved_hosts:
-                    del self._resolved_hosts[key]
-
-        raise aiohttp.ConnectionError('All hosts are unreachable.')
-
-
 class HttpRequest:
 
     GET_METHODS = {'DELETE', 'GET', 'HEAD', 'OPTIONS'}
@@ -382,11 +163,20 @@ class HttpRequest:
     _writer = None  # async task for streaming data
     _continue = None  # waiter future for '100 Continue' response
 
+    # Adding weakref to self for _writer cancelling doesn't make sense:
+    # _writer exists until .write_bytes coro is finished,
+    # .write_bytes generator has strong reference to self and `del request`
+    # doesn't produce request finalization.
+    # After .write_bytes is done _writer has set to None and we have nothing
+    # to cancel.
+    # Maybe we need to add .cancel() method to HttpRequest through for
+    # forced closing request sending.
+
     def __init__(self, method, url, *,
                  params=None, headers=None, data=None, cookies=None,
                  files=None, auth=None, encoding='utf-8', version=(1, 1),
                  compress=None, chunked=None, expect100=False,
-                 verify_ssl=True, loop=None):
+                 verify_ssl=True, loop=None, using_proxy=False):
         self.url = url
         self.method = method.upper()
         self.encoding = encoding
@@ -394,6 +184,7 @@ class HttpRequest:
         self.compress = compress
         self.verify_ssl = verify_ssl
         self.loop = loop
+        self.using_proxy = using_proxy
 
         self.update_version(version)
         self.update_host(url)
@@ -412,12 +203,6 @@ class HttpRequest:
         self.update_transfer_encoding()
         self.update_expect_continue(expect100)
 
-    def __del__(self):
-        """Close request on GC"""
-        if self._writer is not None:
-            self._writer.cancel()
-            self._writer = None
-
     def update_host(self, url):
         """Update destination host, port and connection type (ssl)."""
         scheme, netloc, path, query, fragment = urllib.parse.urlsplit(url)
@@ -439,11 +224,6 @@ class HttpRequest:
 
         # extract host and port
         self.ssl = scheme == 'https'
-        if self.ssl and not self.verify_ssl:
-            sslcontext = self.ssl = ssl.SSLContext(ssl.PROTOCOL_SSLv23)
-            sslcontext.options |= ssl.OP_NO_SSLv2
-            sslcontext.set_default_verify_paths()
-
         if ':' in netloc:
             netloc, port_s = netloc.split(':', 1)
             try:
@@ -457,6 +237,7 @@ class HttpRequest:
             else:
                 self.port = http.client.HTTP_PORT
 
+        self.scheme = scheme
         self.host = netloc
 
     def update_version(self, version):
@@ -480,8 +261,6 @@ class HttpRequest:
         scheme, netloc, path, query, fragment = urllib.parse.urlsplit(self.url)
         if not path:
             path = '/'
-        else:
-            path = urllib.parse.unquote(path)
 
         if isinstance(params, dict):
             params = list(params.items())
@@ -499,8 +278,13 @@ class HttpRequest:
             else:
                 query = params
 
+        if not self.using_proxy:
+            scheme = ''
+            netloc = ''
+
         self.path = urllib.parse.urlunsplit(
-            ('', '', urllib.parse.quote(path), query, fragment))
+            (scheme, netloc,
+             urllib.parse.quote(path, safe='/%'), query, fragment))
 
     def update_headers(self, headers):
         """Update request headers."""
@@ -672,57 +456,61 @@ class HttpRequest:
             self._continue = asyncio.Future(loop=self.loop)
 
     @asyncio.coroutine
-    def write_bytes(self, request, is_stream):
+    def write_bytes(self, request, reader):
         """Support coroutines that yields bytes objects."""
         # 100 response
         if self._continue is not None:
             yield from self._continue
 
-        if is_stream:
-            exc = None
-            value = None
-            stream = self.body
+        try:
+            if inspect.isgenerator(self.body):
+                exc = None
+                value = None
+                stream = self.body
 
-            while True:
-                try:
-                    if exc is not None:
-                        result = stream.throw(exc)
-                    else:
-                        result = stream.send(value)
-                except StopIteration as exc:
-                    if isinstance(exc.value, bytes):
-                        request.write(exc.value)
-                    break
-                except:
-                    self.response.close(True)
-                    raise
-
-                if isinstance(result, asyncio.Future):
-                    exc = None
-                    value = None
+                while True:
                     try:
-                        value = yield result
-                    except Exception as err:
-                        exc = err
-                elif isinstance(result, bytes):
-                    request.write(result)
-                    value = None
-                else:
-                    raise ValueError(
-                        'Bytes object is expected, got: %s.' % type(result))
+                        if exc is not None:
+                            result = stream.throw(exc)
+                        else:
+                            result = stream.send(value)
+                    except StopIteration as exc:
+                        if isinstance(exc.value, bytes):
+                            yield from request.write(exc.value)
+                        break
+                    except:
+                        self.response.close(True)
+                        raise
+
+                    if isinstance(result, asyncio.Future):
+                        exc = None
+                        value = None
+                        try:
+                            value = yield result
+                        except Exception as err:
+                            exc = err
+                    elif isinstance(result, (bytes, bytearray)):
+                        yield from request.write(result)
+                        value = None
+                    else:
+                        raise ValueError(
+                            'Bytes object is expected, got: %s.' %
+                            type(result))
+            else:
+                if isinstance(self.body, bytes):
+                    self.body = (self.body,)
+
+                for chunk in self.body:
+                    request.write(chunk)
+        except Exception as exc:
+            reader.set_exception(exc)
         else:
-            if isinstance(self.body, bytes):
-                self.body = (self.body,)
+            request.write_eof()
 
-            for chunk in self.body:
-                request.write(chunk)
-
-        request.write_eof()
         self._writer = None
 
-    def send(self, transport):
-        request = aiohttp.Request(
-            transport, self.method, self.path, self.version)
+    def send(self, writer, reader):
+        request = aiohttp.Request(writer, self.method, self.path, self.version)
 
         if self.compress:
             request.add_compression_filter(self.compress)
@@ -733,19 +521,8 @@ class HttpRequest:
         request.add_headers(*self.headers.items())
         request.send_headers()
 
-        is_stream = inspect.isgenerator(self.body)
-
-        if is_stream or self._continue is not None:
-            self._writer = asyncio.async(
-                self.write_bytes(request, is_stream), loop=self.loop)
-        else:
-            if isinstance(self.body, bytes):
-                self.body = (self.body,)
-
-            for chunk in self.body:
-                request.write(chunk)
-
-            request.write_eof()
+        self._writer = asyncio.async(
+            self.write_bytes(request, reader), loop=self.loop)
 
         self.response = HttpResponse(
             self.method, self.path, self.host,
@@ -771,12 +548,13 @@ class HttpResponse(http.client.HTTPMessage):
     reason = None   # Reason-Phrase
 
     cookies = None  # Response cookies (Set-Cookie)
+    content = None  # Payload stream
 
-    content = None  # payload stream
-    stream = None   # input stream
-    transport = None  # current transport
-
+    connection = None  # current connection
+    _reader = None     # input stream
     _response_parser = aiohttp.HttpResponseParser()
+    _connection_wr = None  # weakref to self for releasing connection on del
+    _writer_wr = None  # weakref to self for cancelling writer on del
 
     def __init__(self, method, url, host='', *, writer=None, continue100=None):
         super().__init__()
@@ -786,13 +564,10 @@ class HttpResponse(http.client.HTTPMessage):
         self.host = host
         self._content = None
         self._writer = writer
+        if writer is not None:
+            self._writer_wr = weakref.ref(self, lambda wr: writer.cancel())
         self._continue = continue100
 
-    def __del__(self):
-        if self.transport is not None:
-            logging.warn('HttpResponse has to be closed explicitly!')
-            self.close()
-
     def __repr__(self):
         out = io.StringIO()
         print('<HttpResponse({}{}) [{} {}]>'.format(
@@ -802,13 +577,29 @@ class HttpResponse(http.client.HTTPMessage):
 
     __str__ = __repr__
 
-    def start(self, stream, transport):
+    def waiting_for_continue(self):
+        return self._continue is not None
+
+    def _setup_connection(self, connection):
+        self._reader = connection.reader
+        self.connection = connection
+
+        msg = ('HttpResponse has to be closed explicitly! {}:{}:{}'
+               .format(self.method, self.host, self.url))
+
+        def _do_close_connection(wr, connection=connection, msg=msg):
+            warnings.warn(msg, ResourceWarning)
+            connection.close()
+
+        self._connection_wr = weakref.ref(self, _do_close_connection)
+
+    @asyncio.coroutine
+    def start(self, connection, read_until_eof=False):
         """Start response processing."""
-        self.stream = stream
-        self.transport = transport
+        self._setup_connection(connection)
 
         while True:
-            httpstream = stream.set_parser(self._response_parser)
+            httpstream = self._reader.set_parser(self._response_parser)
 
             # read response
             self.message = yield from httpstream.read()
@@ -829,24 +620,32 @@ class HttpResponse(http.client.HTTPMessage):
             self.add_header(hdr, val)
 
         # payload
-        self.content = stream.set_parser(
-            aiohttp.HttpPayloadParser(self.message))
+        self.content = self._reader.set_parser(
+            aiohttp.HttpPayloadParser(self.message, readall=read_until_eof))
 
         # cookies
         self.cookies = http.cookies.SimpleCookie()
         if 'Set-Cookie' in self:
             for hdr in self.get_all('Set-Cookie'):
-                self.cookies.load(hdr)
+                try:
+                    self.cookies.load(hdr)
+                except http.cookies.CookieError as exc:
+                    logging.warn('Can not load response cookies: %s', exc)
 
         return self
 
     def close(self, force=False):
-        if self.transport is not None:
-            self.transport.close(force)
-            self.transport = None
-        if (self._writer is not None) and not self._writer.done():
+        if self.connection is not None:
+            if force:
+                self.connection.close()
+            else:
+                self.connection.release()
+            self.connection = None
+            self._connection_wr = None
+        if self._writer is not None and not self._writer.done():
             self._writer.cancel()
             self._writer = None
+            self._writer_wr = None
 
     @asyncio.coroutine
     def wait_for_close(self):
@@ -855,6 +654,7 @@ class HttpResponse(http.client.HTTPMessage):
                 yield from self._writer
             finally:
                 self._writer = None
+                self._writer_wr = None
         self.close()
 
     @asyncio.coroutine
@@ -889,6 +689,19 @@ class HttpResponse(http.client.HTTPMessage):
 
         return data
 
+    @asyncio.coroutine
+    def read_and_close(self, decode=False):
+        """Read response payload and then close response."""
+        try:
+            payload = yield from self.read(decode)
+        except:
+            self.close(True)
+            raise
+        else:
+            self.close()
+
+        return payload
+
 
 def str_to_bytes(s, encoding='utf-8'):
     if isinstance(s, str):
@@ -949,3 +762,152 @@ def encode_multipart_data(fields, boundary, encoding='utf-8', chunk_size=8196):
             yield b'\r\n'
 
     yield b'--' + boundary + b'--\r\n'
+
+
+class HttpClient:
+    """Allow to use mutiple hosts with same path. And automatically
+    mark failed hosts.
+    """
+
+    def __init__(self, hosts, *, method=None, path=None,
+                 ssl=False, timeout=None,
+                 conn_pool=True, conn_timeout=None, failed_timeout=5.0,
+                 resolve=True, resolve_timeout=360.0, verify_ssl=True,
+                 loop=None):
+        super().__init__()
+
+        if isinstance(hosts, str):
+            hosts = (hosts,)
+
+        if not hosts:
+            raise ValueError('Hosts are required')
+
+        self._hosts = []
+        for host in hosts:
+            if isinstance(host, str):
+                if ':' in host:
+                    host, port = host.split(':')
+                    try:
+                        port = int(port)
+                    except:
+                        raise ValueError('Port has to be integer: %s' % host)
+                else:
+                    port = 80
+            else:
+                host, port = host
+
+            self._hosts.append((host, port))
+
+        self._method = method
+        self._path = path
+        self._timeout = timeout
+        self._schema = 'https' if ssl else 'http'
+
+        self._failed = collections.deque()
+        self._failed_handle = None
+        self._failed_timeout = failed_timeout
+
+        if loop is None:
+            loop = asyncio.get_event_loop()
+        self._loop = loop
+
+        if conn_pool:
+            self._connector = aiohttp.TCPConnector(
+                share_cookies=True, conn_timeout=conn_timeout,
+                resolve=resolve, verify_ssl=verify_ssl, loop=loop)
+
+            self._resolve_timeout = resolve_timeout
+            self._resolve_handle = self._loop.call_later(
+                self._resolve_timeout, self._cleanup_resolved_host)
+        else:
+            self._connector = None
+
+    def _cleanup_resolved_host(self):
+        if self._connector:
+            self._connector.clear_resolved_hosts()
+            self._resolve_handle.cancel()
+            self._resolve_handle = self._loop.call_later(
+                self._resolve_timeout, self._cleanup_resolved_host)
+
+    def _resurrect_failed(self):
+        now = int(time.time())
+
+        while self._failed:
+            if (now - self._failed[0][1]) >= self._failed_timeout:
+                self._hosts.append(self._failed.popleft()[0])
+            else:
+                break
+
+        if self._failed:
+            self._failed_handle = self._loop.call_later(
+                self._failed_timeout, self._resurrect_failed)
+        else:
+            self._failed_handle = None
+
+    @asyncio.coroutine
+    def request(self, method=None, path=None, *,
+                params=None,
+                data=None,
+                headers=None,
+                cookies=None,
+                files=None,
+                auth=None,
+                allow_redirects=True,
+                max_redirects=10,
+                encoding='utf-8',
+                version=(1, 1),
+                compress=None,
+                timeout=None,
+                chunked=None,
+                expect100=False,
+                read_until_eof=True):
+
+        if method is None:
+            method = self._method
+        if path is None:
+            path = self._path
+        if timeout is None:
+            timeout = self._timeout
+
+        # if all hosts marked as failed try first from failed
+        if not self._hosts:
+            self._hosts.append(self._failed.popleft()[0])
+
+        hosts = self._hosts
+
+        while hosts:
+            idx = random.randint(0, len(hosts)-1)
+
+            info = hosts[idx]
+            url = urllib.parse.urljoin(
+                '{}://{}:{}'.format(self._schema, info[0], info[1]), path)
+
+            try:
+                resp = yield from request(
+                    method, url, params=params, data=data, headers=headers,
+                    cookies=cookies, files=files, auth=auth,
+                    encoding=encoding, allow_redirects=allow_redirects,
+                    version=version, max_redirects=max_redirects,
+                    timeout=timeout, compress=compress, chunked=chunked,
+                    expect100=expect100, read_until_eof=read_until_eof,
+                    connector=self._connector, loop=self._loop)
+            except (aiohttp.ConnectionError, aiohttp.TimeoutError):
+                pass
+            else:
+                if 500 <= resp.status <= 600:
+                    self._cleanup_resolved_host()
+
+                return resp
+
+            if info in hosts:
+                # could be removed concurrently
+                hosts.remove(info)
+                self._failed.append((info, int(time.time())))
+                if not self._failed_handle:
+                    self._failed_handle = self._loop.call_later(
+                        self._failed_timeout, self._resurrect_failed)
+
+                if self._connector:
+                    self._connector.clear_resolved_hosts(info[0], info[1])
+
+        raise aiohttp.ConnectionError('All hosts are unreachable.')
