@@ -59,9 +59,13 @@ __all__ = ['EofStream', 'StreamParser', 'StreamProtocol',
            'ParserBuffer', 'DataQueue', 'LinesParser', 'ChunksParser']
 
 import asyncio
+import asyncio.streams
 import collections
 import inspect
 
+BUF_LIMIT = 2**14
+DEFAULT_LIMIT = 2**16
+
 
 class EofStream(Exception):
     """eof stream indication."""
@@ -78,16 +82,41 @@ class StreamParser:
     unset_parser() sends EofStream into parser and then removes it.
     """
 
-    def __init__(self, *, loop=None, inbuf=None):
+    def __init__(self, *, loop=None, buf=None,
+                 paused=True, limit=DEFAULT_LIMIT):
         self._loop = loop
         self._eof = False
         self._exception = None
         self._parser = None
+        self._transport = None
+        self._limit = limit
+        self._paused = False
+        self._stream_paused = paused
         self._output = None
-        self._input = inbuf if inbuf is not None else ParserBuffer()
+        self._buffer = buf if buf is not None else ParserBuffer()
 
-    def is_connected(self):
-        return not self._eof
+    @property
+    def output(self):
+        return self._output
+
+    def set_transport(self, transport):
+        assert self._transport is None, 'Transport already set'
+        self._transport = transport
+
+    def at_eof(self):
+        return self._eof
+
+    def pause_stream(self):
+        self._stream_paused = True
+
+    def resume_stream(self):
+        if self._paused and self._buffer.size <= self._limit:
+            self._paused = False
+            self._transport.resume_reading()
+
+        self._stream_paused = False
+        if self._parser and self._buffer:
+            self.feed_data(b'')
 
     def exception(self):
         return self._exception
@@ -102,10 +131,10 @@ class StreamParser:
 
     def feed_data(self, data):
         """send data to current parser or store in buffer."""
-        if not data:
+        if data is None:
             return
 
-        if self._parser:
+        if self._parser and not self._stream_paused:
             try:
                 self._parser.send(data)
             except StopIteration:
@@ -117,12 +146,26 @@ class StreamParser:
                 self._output = None
                 self._parser = None
         else:
-            self._input.feed_data(data)
+            self._buffer.feed_data(data)
+
+        if (self._transport is not None and not self._paused and
+                self._buffer.size > 2*self._limit):
+            try:
+                self._transport.pause_reading()
+            except NotImplementedError:
+                # The transport can't be paused.
+                # We'll just have to buffer all data.
+                # Forget the transport so we don't keep trying.
+                self._transport = None
+            else:
+                self._paused = True
 
     def feed_eof(self):
         """send eof to all parsers, recursively."""
         if self._parser:
             try:
+                if self._buffer:
+                    self._parser.send(b'')
                 self._parser.throw(EofStream())
             except StopIteration:
                 pass
@@ -134,6 +177,7 @@ class StreamParser:
             self._parser = None
             self._output = None
 
+        self._buffer.shrink()
         self._eof = True
 
     def set_parser(self, parser):
@@ -141,13 +185,13 @@ class StreamParser:
         if self._parser:
             self.unset_parser()
 
-        output = DataQueue(loop=self._loop)
+        output = DataQueue(self, loop=self._loop)
         if self._exception:
             output.set_exception(self._exception)
             return output
 
         # init parser
-        p = parser(output, self._input)
+        p = parser(output, self._buffer)
         assert inspect.isgenerator(p), 'Generator is required'
 
         try:
@@ -169,6 +213,9 @@ class StreamParser:
 
     def unset_parser(self):
         """unset parser, send eof to the parser and then remove it."""
+        if self._buffer:
+            self._buffer.shrink()
+
         if self._parser is None:
             return
 
@@ -185,37 +232,65 @@ class StreamParser:
             self._parser = None
 
 
-class StreamProtocol(StreamParser, asyncio.Protocol):
-    """asyncio's stream protocol based on StreamParser"""
+class StreamProtocol(asyncio.streams.FlowControlMixin, asyncio.Protocol):
+    """Helper class to adapt between Protocol and StreamReader."""
 
-    transport = None
+    def __init__(self, *, loop=None, **kwargs):
+        super().__init__(loop=loop)
 
-    data_received = StreamParser.feed_data
+        self.transport = None
+        self.writer = None
+        self.reader = StreamParser(loop=loop, **kwargs)
 
-    eof_received = StreamParser.feed_eof
+    def is_connected(self):
+        return self.transport is not None
 
     def connection_made(self, transport):
         self.transport = transport
+        self.reader.set_transport(transport)
+        self.writer = asyncio.streams.StreamWriter(
+            transport, self, self.reader, self._loop)
 
     def connection_lost(self, exc):
         self.transport = None
 
-        if exc is not None:
-            self.set_exception(exc)
+        if exc is None:
+            self.reader.feed_eof()
         else:
-            self.feed_eof()
+            self.reader.set_exception(exc)
+
+        super().connection_lost(exc)
+
+    def data_received(self, data):
+        self.reader.feed_data(data)
+
+    def eof_received(self):
+        self.reader.feed_eof()
+
+    def _make_drain_waiter(self):
+        if not self._paused:
+            return ()
+        waiter = self._drain_waiter
+        if waiter is None or waiter.cancelled():
+            waiter = asyncio.Future(loop=self._loop)
+            self._drain_waiter = waiter
+        return waiter
 
 
 class DataQueue:
     """DataQueue is a destination for parsed data."""
 
-    def __init__(self, *, loop=None):
+    def __init__(self, stream, *, loop=None):
+        self._stream = stream
         self._loop = loop
         self._buffer = collections.deque()
         self._eof = False
         self._waiter = None
         self._exception = None
 
+    def at_eof(self):
+        return self._eof
+
     def exception(self):
         return self._exception
 
@@ -251,15 +326,19 @@ class DataQueue:
         if self._exception is not None:
             raise self._exception
 
-        if not self._buffer and not self._eof:
-            assert not self._waiter
-            self._waiter = asyncio.Future(loop=self._loop)
-            yield from self._waiter
+        self._stream.resume_stream()
+        try:
+            if not self._buffer and not self._eof:
+                assert not self._waiter
+                self._waiter = asyncio.Future(loop=self._loop)
+                yield from self._waiter
 
-        if self._buffer:
-            return self._buffer.popleft()
-        else:
-            raise EofStream
+            if self._buffer:
+                return self._buffer.popleft()
+            else:
+                raise EofStream
+        finally:
+            self._stream.pause_stream()
 
 
 class ParserBuffer(bytearray):
@@ -268,15 +347,23 @@ class ParserBuffer(bytearray):
     ParserBuffer provides helper methods for parsers.
     """
 
-    def __init__(self, *args):
+    def __init__(self, *args, limit=BUF_LIMIT):
         super().__init__(*args)
 
         self.offset = 0
         self.size = 0
+        self._limit = limit
+        self._exception = None
         self._writer = self._feed_data()
         next(self._writer)
 
-    def _shrink(self):
+    def exception(self):
+        return self._exception
+
+    def set_exception(self, exc):
+        self._exception = exc
+
+    def shrink(self):
         if self.offset:
             del self[:self.offset]
             self.offset = 0
@@ -290,9 +377,14 @@ class ParserBuffer(bytearray):
                 self.size += chunk_len
                 self.extend(chunk)
 
-                # shrink buffer
-                if (self.offset and len(self) > 8196):
-                    self._shrink()
+            # shrink buffer
+            if (self.offset and len(self) > self._limit):
+                self.shrink()
+
+            if self._exception:
+                self._writer = self._feed_data()
+                next(self._writer)
+                raise self._exception
 
     def feed_data(self, data):
         self._writer.send(data)
@@ -349,6 +441,38 @@ class ParserBuffer(bytearray):
 
             self._writer.send((yield))
 
+    def wait(self, size):
+        """wait() waits for specified amount of bytes
+        then returns data without changing internal buffer."""
+
+        while True:
+            if self.size >= size:
+                return self[self.offset:self.offset + size]
+
+            self._writer.send((yield))
+
+    def waituntil(self, stop, limit=None, exc=ValueError):
+        """waituntil() reads until `stop` bytes sequence."""
+        assert isinstance(stop, bytes) and stop, \
+            'bytes is required: {!r}'.format(stop)
+
+        stop_len = len(stop)
+
+        while True:
+            pos = self.find(stop, self.offset)
+            if pos >= 0:
+                end = pos + stop_len
+                size = end - self.offset
+                if limit is not None and size > limit:
+                    raise exc('Line is too long. %s' % bytes(self))
+
+                return self[self.offset:end]
+            else:
+                if limit is not None and self.size > limit:
+                    raise exc('Line is too long. %s' % bytes(self))
+
+            self._writer.send((yield))
+
     def skip(self, size):
         """skip() skips specified amount of bytes."""
 
