@@ -2,12 +2,14 @@
 
 __all__ = ['HttpMessage', 'Request', 'Response',
            'RawRequestMessage', 'RawResponseMessage',
-           'HttpRequestParser', 'HttpResponseParser', 'HttpPayloadParser']
+           'HttpPrefixParser', 'HttpRequestParser', 'HttpResponseParser',
+           'HttpPayloadParser']
 
 import collections
 import functools
 import http.server
 import itertools
+import logging
 import re
 import sys
 import zlib
@@ -115,6 +117,34 @@ class HttpParser:
         return headers, close_conn, encoding
 
 
+class HttpPrefixParser:
+    """Waits for 'HTTP' prefix (non destructive)"""
+
+    def __init__(self, allowed_methods=()):
+        self.allowed_methods = [m.upper() for m in allowed_methods]
+
+    def __call__(self, out, buf):
+        try:
+            raw_data = yield from buf.waituntil(b' ', 24)
+            method = raw_data.decode('ascii', 'surrogateescape').strip()
+
+            # method
+            method = method.upper()
+            if not METHRE.match(method):
+                raise errors.BadStatusLine(method)
+
+            # allowed method
+            if self.allowed_methods and method not in self.allowed_methods:
+                raise errors.HttpMethodNotAllowed(method)
+
+            out.feed_data(method)
+            out.feed_eof()
+        except aiohttp.EofStream:
+            # Presumably, the server closed the connection before
+            # sending a valid response.
+            pass
+
+
 class HttpRequestParser(HttpParser):
     """Read request status line. Exception errors.BadStatusLine
     could be raised in case of any errors in status line.
@@ -261,8 +291,11 @@ class HttpPayloadParser:
             elif length > 0:
                 yield from self.parse_length_payload(out, buf, length)
         else:
-            if self.readall:
+            if self.readall and getattr(self.message, 'code', 0) != 204:
                 yield from self.parse_eof_payload(out, buf)
+            elif getattr(self.message, 'method', None) in ('PUT', 'POST'):
+                logging.warn(  # pragma: no cover
+                    'Content-Length or Transfer-Encoding header is required')
 
         out.feed_eof()
 
@@ -477,8 +510,8 @@ class HttpMessage:
 
     HOP_HEADERS = None  # Must be set by subclass.
 
-    SERVER_SOFTWARE = 'Python/{0[0]}.{0[1]} asyncio/0.1'.format(
-        sys.version_info)
+    SERVER_SOFTWARE = 'Python/{0[0]}.{0[1]} aiohttp/{1}'.format(
+        sys.version_info, aiohttp.__version__)
 
     status = None
     status_line = b''
@@ -507,6 +540,7 @@ class HttpMessage:
         self.headers = collections.deque()
         self.headers_sent = False
         self.output_length = 0
+        self._output_size = 0
 
     def force_close(self):
         self.closing = True
@@ -619,10 +653,13 @@ class HttpMessage:
         """write() writes chunk of data to a steram by using different writers.
         writer uses filter to modify chunk of data. write_eof() indicates
         end of stream. writer can't be used after write_eof() method
-        being called."""
+        being called. write() return drain future.
+        """
         assert (isinstance(chunk, (bytes, bytearray)) or
                 chunk is EOF_MARKER), chunk
 
+        size = self.output_length
+
         if self._send_headers and not self.headers_sent:
             self.send_headers()
 
@@ -637,6 +674,14 @@ class HttpMessage:
             if chunk is not EOF_MARKER:
                 self.writer.send(chunk)
 
+        self._output_size += self.output_length - size
+
+        if self._output_size > 64 * 1024:
+            self._output_size = 0
+            return self.transport.drain()
+        else:
+            return ()
+
     def write_eof(self):
         self.write(EOF_MARKER)
         try:
@@ -644,6 +689,8 @@ class HttpMessage:
         except StopIteration:
             pass
 
+        return self.transport.drain()
+
     def _write_chunked_payload(self):
         """Write data in chunked transfer encoding."""
         while True:
