@@ -4,10 +4,10 @@ __all__ = ['ServerHttpProtocol']
 
 import asyncio
 import http.server
-import inspect
 import logging
 import time
 import traceback
+import socket
 
 import aiohttp
 from aiohttp import errors, utils
@@ -30,7 +30,7 @@ ACCESS_LOG_FORMAT = (
     '%(h)s %(l)s %(u)s %(t)s "%(r)s" %(s)s %(b)s "%(f)s" "%(a)s"')
 
 
-class ServerHttpProtocol(asyncio.Protocol):
+class ServerHttpProtocol(aiohttp.StreamProtocol):
     """Simple http protocol implementation.
 
     ServerHttpProtocol handles incoming http request. It reads request line,
@@ -44,55 +44,61 @@ class ServerHttpProtocol(asyncio.Protocol):
     log: custom logging object
     debug: enable debug mode
     keep_alive: number of seconds before closing keep alive connection
+    timeout: slow request timeout
     loop: event loop object
     """
     _request_count = 0
     _request_handler = None
+    _reading_request = False
     _keep_alive = False  # keep transport open
     _keep_alive_handle = None  # keep alive timer handle
+    _timeout_handle = None  # slow request timer handle
 
     _request_parser = aiohttp.HttpRequestParser()  # default request parser
 
-    def __init__(self, *, loop=None,
-                 keep_alive=None, debug=False, log=logging,
-                 access_log=ACCESS_LOG, access_log_format=ACCESS_LOG_FORMAT):
-        self._keep_alive_period = keep_alive  # number of seconds to keep alive
+    def __init__(self, *, loop=None, keep_alive=None,
+                 timeout=15, tcp_keepalive=True, allowed_methods=(),
+                 debug=False, log=logging, access_log=ACCESS_LOG,
+                 access_log_format=ACCESS_LOG_FORMAT, **kwargs):
+        super().__init__(loop=loop, **kwargs)
 
-        if keep_alive and loop is None:
-            loop = asyncio.get_event_loop()
-        self._loop = loop
+        self._keep_alive_period = keep_alive  # number of seconds to keep alive
+        self._timeout = timeout  # slow request timeout
+        self._tcp_keepalive = tcp_keepalive  # use detection of broken socket
+        self._request_prefix = aiohttp.HttpPrefixParser(allowed_methods)
+        self._loop = loop if loop is not None else asyncio.get_event_loop()
 
         self.log = log
         self.debug = debug
         self.access_log = access_log
         self.access_log_format = access_log_format
-        self.transport = None
 
     def closing(self):
         """Worker process is about to exit, we need cleanup everything and
         stop accepting requests. It is especially important for keep-alive
         connections."""
         self._keep_alive = False
-        if self._keep_alive_handle is not None:
-            self._keep_alive_handle.cancel()
-            self._keep_alive_handle = None
 
-        if self._request_handler is None and self.transport is not None:
+        if ((not self._reading_request or self._request_handler is None) and
+                self.transport is not None):
             self.transport.close()
 
     def connection_made(self, transport):
-        self.transport = transport
-        self.stream = aiohttp.StreamParser(loop=self._loop)
+        super().connection_made(transport)
+
+        if self._tcp_keepalive and hasattr(socket, 'SO_KEEPALIVE'):
+            sock = transport.get_extra_info('socket')
+            sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
+
         self._request_handler = asyncio.async(self.start(), loop=self._loop)
 
-    def data_received(self, data):
-        self.stream.feed_data(data)
-
-    def eof_received(self):
-        self.stream.feed_eof()
+        # start slow request timer
+        if self._timeout:
+            self._timeout_handle = self._loop.call_later(
+                self._timeout, self.cancel_slow_request)
 
     def connection_lost(self, exc):
-        self.stream.feed_eof()
+        super().connection_lost(exc)
 
         if self._request_handler is not None:
             self._request_handler.cancel()
@@ -100,6 +106,9 @@ class ServerHttpProtocol(asyncio.Protocol):
         if self._keep_alive_handle is not None:
             self._keep_alive_handle.cancel()
             self._keep_alive_handle = None
+        if self._timeout_handle is not None:
+            self._timeout_handle.cancel()
+            self._timeout_handle = None
 
     def keep_alive(self, val):
         self._keep_alive = val
@@ -121,6 +130,16 @@ class ServerHttpProtocol(asyncio.Protocol):
     def log_exception(self, *args, **kw):
         self.log.exception(*args, **kw)
 
+    def cancel_slow_request(self):
+        if self._request_handler is not None:
+            self._request_handler.cancel()
+            self._request_handler = None
+
+        if self.transport is not None:
+            self.transport.close()
+
+        self.log_debug('Close slow request.')
+
     @asyncio.coroutine
     def start(self):
         """Start processing of incoming requests.
@@ -130,38 +149,61 @@ class ServerHttpProtocol(asyncio.Protocol):
         or response handling. Connection is being closed always unless
         keep_alive(True) specified.
         """
+        reader = self.reader
 
         while True:
-            info = None
             message = None
-            self._request_count += 1
             self._keep_alive = False
+            self._request_count += 1
+            self._reading_request = False
 
             try:
-                httpstream = self.stream.set_parser(self._request_parser)
-                message = yield from httpstream.read()
+                prefix = reader.set_parser(self._request_prefix)
+                yield from prefix.read()
+                self._reading_request = True
 
-                # cancel keep-alive timer
+                # stop keep-alive timer
                 if self._keep_alive_handle is not None:
                     self._keep_alive_handle.cancel()
                     self._keep_alive_handle = None
 
-                payload = self.stream.set_parser(
-                    aiohttp.HttpPayloadParser(message))
+                # start slow request timer
+                if self._timeout and self._timeout_handle is None:
+                    self._timeout_handle = self._loop.call_later(
+                        self._timeout, self.cancel_slow_request)
+
+                # read request headers
+                httpstream = reader.set_parser(self._request_parser)
+                message = yield from httpstream.read()
+
+                # cancel slow request timer
+                if self._timeout_handle is not None:
+                    self._timeout_handle.cancel()
+                    self._timeout_handle = None
+
+                payload = reader.set_parser(aiohttp.HttpPayloadParser(message))
 
                 handler = self.handle_request(message, payload)
-                if (inspect.isgenerator(handler) or
+                if (asyncio.iscoroutine(handler) or
                         isinstance(handler, asyncio.Future)):
                     yield from handler
 
-            except asyncio.CancelledError:
+            except (asyncio.CancelledError, ConnectionError):
                 self.log_debug('Ignored premature client disconnection.')
                 break
             except errors.HttpException as exc:
-                self.handle_error(exc.code, info, message, exc, exc.headers)
+                self.handle_error(exc.code, message, None, exc, exc.headers)
             except Exception as exc:
-                self.handle_error(500, info, message, exc)
+                self.handle_error(500, message, None, exc)
             finally:
+                if reader.output and not reader.output.at_eof():
+                    self.log_debug('Uncompleted request.')
+                    self._request_handler = None
+                    self.transport.close()
+                    break
+                else:
+                    reader.unset_parser()
+
                 if self._request_handler:
                     if self._keep_alive and self._keep_alive_period:
                         self.log_debug(
@@ -171,8 +213,8 @@ class ServerHttpProtocol(asyncio.Protocol):
                             self._keep_alive_period, self.transport.close)
                     else:
                         self.log_debug('Close client connection.')
-                        self.transport.close()
                         self._request_handler = None
+                        self.transport.close()
                         break
                 else:
                     break
@@ -208,7 +250,7 @@ class ServerHttpProtocol(asyncio.Protocol):
             html = DEFAULT_ERROR_MESSAGE.format(
                 status=status, reason=reason, message=msg)
 
-            response = aiohttp.Response(self.transport, status, close=True)
+            response = aiohttp.Response(self.writer, status, close=True)
             response.add_headers(
                 ('Content-Type', 'text/html'),
                 ('Content-Length', str(len(html))))
@@ -234,7 +276,7 @@ class ServerHttpProtocol(asyncio.Protocol):
         """
         now = time.time()
         response = aiohttp.Response(
-            self.transport, 404, http_version=message.version, close=True)
+            self.writer, 404, http_version=message.version, close=True)
 
         body = b'Page Not Found!'
 
@@ -243,7 +285,9 @@ class ServerHttpProtocol(asyncio.Protocol):
             ('Content-Length', str(len(body))))
         response.send_headers()
         response.write(body)
-        response.write_eof()
+        drain = response.write_eof()
 
         self.keep_alive(False)
         self.log_access(message, None, response, time.time() - now)
+
+        return drain
