@@ -31,16 +31,22 @@ def run_briefly(loop):
 
 
 @contextlib.contextmanager
-def run_server(loop, *, host='127.0.0.1', port=0, use_ssl=False, router=None):
+def run_server(loop, *, listen_addr=('127.0.0.1', 0),
+               use_ssl=False, router=None):
     properties = {}
     transports = []
 
-    class HttpServer:
+    class HttpRequestHandler:
 
-        def __init__(self, host, port):
-            self.host = host
-            self.port = port
-            self.address = (host, port)
+        def __init__(self, addr):
+            if isinstance(addr, tuple):
+                host, port = addr
+                self.host = host
+                self.port = port
+            else:
+                self.host = host = 'localhost'
+                self.port = port = 0
+            self.address = addr
             self._url = '{}://{}:{}'.format(
                 'https' if use_ssl else 'http', host, port)
 
@@ -58,6 +64,7 @@ def run_server(loop, *, host='127.0.0.1', port=0, use_ssl=False, router=None):
 
         def connection_made(self, transport):
             transports.append(transport)
+
             super().connection_made(transport)
 
         def handle_request(self, message, payload):
@@ -81,13 +88,11 @@ def run_server(loop, *, host='127.0.0.1', port=0, use_ssl=False, router=None):
                     pass
 
                 rob = router(
-                    self, properties,
-                    self.transport, message, bytes(body))
+                    self, properties, self.transport, message, bytes(body))
                 rob.dispatch()
 
             else:
-                response = aiohttp.Response(
-                    self.transport, 200, message.version)
+                response = aiohttp.Response(self.writer, 200, message.version)
 
                 text = b'Test message'
                 response.add_header('Content-type', 'text/plain')
@@ -109,10 +114,20 @@ def run_server(loop, *, host='127.0.0.1', port=0, use_ssl=False, router=None):
         thread_loop = asyncio.new_event_loop()
         asyncio.set_event_loop(thread_loop)
 
-        server = thread_loop.run_until_complete(
-            thread_loop.create_server(
+        if isinstance(listen_addr, tuple):
+            host, port = listen_addr
+            server_coroutine = thread_loop.create_server(
                 lambda: TestHttpServer(keep_alive=0.5),
-                host, port, ssl=sslcontext))
+                host, port, ssl=sslcontext)
+        else:
+            try:
+                os.unlink(listen_addr)
+            except FileNotFoundError:
+                pass
+            server_coroutine = thread_loop.create_unix_server(
+                lambda: TestHttpServer(keep_alive=0.5),
+                listen_addr, ssl=sslcontext)
+        server = thread_loop.run_until_complete(server_coroutine)
 
         waiter = asyncio.Future(loop=thread_loop)
         loop.call_soon_threadsafe(
@@ -142,7 +157,7 @@ def run_server(loop, *, host='127.0.0.1', port=0, use_ssl=False, router=None):
 
     thread_loop, waiter, addr = loop.run_until_complete(fut)
     try:
-        yield HttpServer(*addr)
+        yield HttpRequestHandler(addr)
     finally:
         thread_loop.call_soon_threadsafe(waiter.set_result, None)
         server_thread.join()
@@ -198,9 +213,10 @@ class Router:
         return self._response(self._start_response(404))
 
     def _start_response(self, code):
-        return aiohttp.Response(self._transport, code)
+        return aiohttp.Response(self._srv.writer, code)
 
-    def _response(self, response, body=None, headers=None, chunked=False):
+    def _response(self, response, body=None,
+                  headers=None, chunked=False, write_body=None):
         r_headers = {}
         for key, val in self._headers.items():
             key = '-'.join(p.capitalize() for p in key.split('-'))
@@ -279,7 +295,14 @@ class Router:
         response.send_headers()
 
         # write payload
-        response.write(client.str_to_bytes(body))
+        if write_body:
+            try:
+                write_body(response, body)
+            except:
+                return
+        else:
+            response.write(client.str_to_bytes(body))
+
         response.write_eof()
 
         # keep-alive
