@@ -139,8 +139,8 @@ def parse_message(buf):
 
 class WebSocketWriter:
 
-    def __init__(self, transport):
-        self.transport = transport
+    def __init__(self, writer):
+        self.writer = writer
 
     def _send_frame(self, message, opcode):
         """Send a frame over the websocket with message as its payload."""
@@ -154,7 +154,7 @@ class WebSocketWriter:
         else:
             header += bytes([127]) + struct.pack('!Q', msg_length)
 
-        self.transport.write(header + message)
+        self.writer.write(header + message)
 
     def pong(self):
         """Send pong message."""
@@ -194,30 +194,30 @@ def do_handshake(method, headers, transport):
     headers = dict(((hdr, val) for hdr, val in headers if hdr in WS_HDRS))
 
     if 'websocket' != headers.get('UPGRADE', '').lower().strip():
-        raise errors.BadRequestException(
+        raise errors.HttpBadRequest(
             'No WebSocket UPGRADE hdr: {}\n'
             'Can "Upgrade" only to "WebSocket".'.format(
                 headers.get('UPGRADE')))
 
     if 'upgrade' not in headers.get('CONNECTION', '').lower():
-        raise errors.BadRequestException(
+        raise errors.HttpBadRequest(
             'No CONNECTION upgrade hdr: {}'.format(
                 headers.get('CONNECTION')))
 
     # check supported version
     version = headers.get('SEC-WEBSOCKET-VERSION')
     if version not in ('13', '8', '7'):
-        raise errors.BadRequestException(
+        raise errors.HttpBadRequest(
             'Unsupported version: {}'.format(version))
 
     # check client handshake for validity
     key = headers.get('SEC-WEBSOCKET-KEY')
     try:
         if not key or len(base64.b64decode(key)) != 16:
-            raise errors.BadRequestException(
+            raise errors.HttpBadRequest(
                 'Handshake error: {!r}'.format(key))
     except binascii.Error:
-        raise errors.BadRequestException(
+        raise errors.HttpBadRequest(
             'Handshake error: {!r}'.format(key)) from None
 
     # response code, headers, parser, writer
