@@ -25,7 +25,7 @@ ARGS.add_argument(
     default=2, type=int, help='Number of workers.')
 
 
-class HttpServer(aiohttp.server.ServerHttpProtocol):
+class HttpRequestHandler(aiohttp.server.ServerHttpProtocol):
 
     @asyncio.coroutine
     def handle_request(self, message, payload):
@@ -55,7 +55,8 @@ class HttpServer(aiohttp.server.ServerHttpProtocol):
             raise aiohttp.HttpErrorException(
                 302, headers=(('URI', path), ('Location', path)))
 
-        response = aiohttp.Response(self.transport, 200)
+        response = aiohttp.Response(
+            self.writer, 200, http_version=message.version)
         response.add_header('Transfer-Encoding', 'chunked')
 
         # content encoding
@@ -125,7 +126,8 @@ class ChildProcess:
         loop.add_signal_handler(signal.SIGINT, stop)
 
         f = loop.create_server(
-            lambda: HttpServer(debug=True, keep_alive=75), sock=self.sock)
+            lambda: HttpRequestHandler(debug=True, keep_alive=75),
+            sock=self.sock)
         srv = loop.run_until_complete(f)
         x = srv.sockets[0]
         print('Starting srv worker process {} on {}'.format(
@@ -145,7 +147,7 @@ class ChildProcess:
         write_transport, _ = yield from self.loop.connect_write_pipe(
             aiohttp.StreamProtocol, os.fdopen(self.down_write, 'wb'))
 
-        reader = read_proto.set_parser(websocket.WebSocketParser)
+        reader = read_proto.reader.set_parser(websocket.WebSocketParser)
         writer = websocket.WebSocketWriter(write_transport)
 
         while True:
@@ -239,7 +241,7 @@ class Worker:
             aiohttp.StreamProtocol, os.fdopen(up_write, 'wb'))
 
         # websocket protocol
-        reader = proto.set_parser(websocket.WebSocketParser)
+        reader = proto.reader.set_parser(websocket.WebSocketParser)
         writer = websocket.WebSocketWriter(write_transport)
 
         # store info
